/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/tcp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "tcpjunk.h"
#include "gui.h"
#include "utils.h"

#include <linux/types.h>

struct in6_ifreq
  {
	struct in6_addr ifr6_addr;
	__u32 ifr6_prefixlen;
	int ifr6_ifindex;
  };

void show_peer_addr(struct tj_data *inst_data, int cl_socket)
  {
  /* print destination IP address */
  char addr[128];
  if (inst_data->ipv6)
	{
	struct sockaddr_in6 const *sin;
	sin = (struct sockaddr_in6 *) inst_data->remote_addr->ai_addr;
	sprintf(log_msg, "connected to: %s", inet_ntop(
		inst_data->remote_addr->ai_family, &sin->sin6_addr, addr, sizeof addr));
	}
  else
	{
	struct sockaddr_in const *sin;
	sin = (struct sockaddr_in *) inst_data->remote_addr->ai_addr;
	sprintf(log_msg, "connected to: %s", inet_ntop(
		inst_data->remote_addr->ai_family, &sin->sin_addr, addr, sizeof addr));
	}
  tj_log(inst_data, log_msg, c_black);
  }

char *addr_ntop(const struct sockaddr *sa, socklen_t salen)
  {
  char port[8];
  static char addr[128];

  if (sa->sa_family == AF_INET)
	{
	struct sockaddr_in *sin = (struct sockaddr_in *) sa;
	if (inet_ntop(AF_INET, &sin->sin_addr, addr, sizeof(addr)) == NULL)
	  return (NULL);
	if (ntohs(sin->sin_port) != 0)
	  {
	  snprintf(port, sizeof(port), ":%d", ntohs(sin->sin_port));
	  strcat(addr, port);
	  }
	}
  else
	{
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) sa;
	if (inet_ntop(AF_INET6, &sin6->sin6_addr, addr, sizeof(addr)) == NULL)
	  return (NULL);

	if (ntohs(sin6->sin6_port) != 0)
	  {
	  snprintf(port, sizeof(port), ":%d", ntohs(sin6->sin6_port));
	  strcat(addr, port);
	  }
	}
  return (addr);
  }

int connect_tcp(struct tj_data *inst_data, int sock)
  {
  int rslt = 0, sock_opts;
  int so_ret, get_ret;
  unsigned int get_len = sizeof(get_ret);

  /* set non-blocking on socket */
  if ((sock_opts = fcntl(sock, F_GETFL, 0)) >= 0)
	rslt = fcntl(sock, F_SETFL, sock_opts | O_NONBLOCK);
  if (rslt < 0)
	return -1;

  /* buffer size set */
  if (inst_data->send_size)
	{
	int flag = 1;
	rslt = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &flag,
		sizeof(int));
	if (rslt < 0)
	  {
	  sprintf(log_msg, "setting TCP_NODELAY failed: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  }
	}

  /* max segment size   */
  if (inst_data->max_seg)
	{
	rslt = setsockopt(sock, SOL_TCP, TCP_MAXSEG, (char*) &inst_data->max_seg,
		sizeof(inst_data->max_seg));
	if (rslt < 0)
	  {
	  sprintf(log_msg, "setting TCP_MAXSEG failed: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  }
	}

  /* connect */
  rslt = connect(sock, inst_data->remote_addr->ai_addr,
	  inst_data->remote_addr->ai_addrlen);
  if ((rslt < 0) && (errno != EINPROGRESS))
	return -1;

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timout */
  struct timeval timeout;
  timeout.tv_sec = inst_data->conn_timeout;
  timeout.tv_usec = 0;

  rslt = select(sock + 1, NULL, &set, NULL, &timeout);
  if (rslt > 0)
	{
	/* handle errors */
	so_ret = getsockopt(sock, SOL_SOCKET, SO_ERROR, &get_ret, &get_len);
	if (so_ret < 0)
	  {
	  sprintf(log_msg, "system request failed: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  return -1;
	  }
	if (get_ret > 0)
	  {
	  errno = get_ret;
	  return -1;
	  }
	}
  else if (rslt <= 0)
	{
	errno = ETIMEDOUT;
	return -1;
	}

  /* all was good, restore socket options */
  rslt = fcntl(sock, F_SETFL, sock_opts);

  if (inst_data->verbose == 1)
	{
	show_peer_addr(inst_data, sock);
	}

  return 0;
  }

void disconnect_tcp(int sock, struct tj_data *inst_data, SSL *ssl, SSL_CTX *ctx)
  {
  int err = 0;
  if (inst_data->use_ssl == 1)
	{
	if (ssl != NULL)
	  err = SSL_clear(ssl);
	if ((err) == -1)
	  {
	  sprintf(log_msg, "SSL Clear %s", ERR_error_string(ERR_get_error(), NULL));
	  tj_log(inst_data, log_msg, c_red);
	  }
	}

  /* close socket */
  err = close(sock);
  if (err < 0)
	{
	sprintf(log_msg, "closing socket failed: %s", strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	}
  else if (inst_data->verbose == 1)
	tj_log(inst_data, "connection closed", c_black);
  }

int read_reply_tcp(int sock, SSL *ssl, struct tj_data *inst_data)
  {
  int sel_rslt, read_rslt;
  int read_success = 0;
  char read_buf[SOCK_BUF_LEN];

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = inst_data->reply_timeout;

  while ((sel_rslt = select(sock + 1, &set, NULL, NULL, &timeout)) > 0
	  || (inst_data->use_ssl && SSL_pending(ssl)))
	{
	/* clean buffer */
	memset(read_buf, 0, sizeof(read_buf));

	/* read clear text */
	if (inst_data->use_ssl == 0)
	  {
	  read_rslt = recv(sock, read_buf, SOCK_BUF_LEN, MSG_DONTWAIT);
	  if (read_rslt < 0)
		{
		sprintf(log_msg, "error reading data from socket: %s", strerror(errno));
		tj_log(inst_data, log_msg, c_red);
		break;
		}
	  }
	/* read using SSL */
	else
	  {
	  read_rslt = SSL_read(ssl, read_buf, SOCK_BUF_LEN);
	  if (read_rslt < 0)
		{
		sprintf(log_msg, "SSL read %s", ERR_error_string(ERR_get_error(), NULL));
		tj_log(inst_data, log_msg, c_red);
		break;
		}
	  }

	/* if no data break */
	if (read_rslt == 0)
	  break;

	/* print reply */
	if (inst_data->debug > 0 && inst_data->gui == 0)
	  {
	  if (inst_data->hexdump)
		{
		printf("\n%s", hex_dump(read_buf, read_rslt));
		}
	  else
		fwrite(read_buf, sizeof(char), read_rslt, stdout);
	  }
	else if (inst_data->gui == 1)
	  gui_show_traffic(read_buf, read_rslt, 1, inst_data);

	/* verbose messages */
	if (inst_data->verbose)
	  {
	  sprintf(log_msg, "received %d bytes", read_rslt);
	  tj_log(inst_data, log_msg, c_blue_hl);
	  }

	if (!read_success)
	  read_success = 1;

	/* reset timeouts */
	timeout.tv_sec = 0;
	timeout.tv_usec = inst_data->reply_timeout;

	}

  if (read_success)
	return 0;
  else
	return 1;

  }

static void show_send_data(struct tj_data *inst_data, char *send_buf,
	int send_buf_len)
  {
  /* print sent buffer */
  if (inst_data->debug > 0 && inst_data->gui == 0)
	{
	if (inst_data->hexdump)
	  printf("\n%s", hex_dump(send_buf, send_buf_len));
	else
	  fwrite(send_buf, sizeof(char), send_buf_len, stdout);
	}
  else if (inst_data->gui == 1)
	gui_show_traffic(send_buf, send_buf_len, 0, inst_data);

  /* print sent buffer size */
  if (inst_data->verbose)
	{
	sprintf(log_msg, "sent %d bytes", send_buf_len);
	tj_log(inst_data, log_msg, c_red_hl);
	}
  }

void send_data_tcp(int sock, SSL *ssl, struct tj_data *inst_data,
	char *send_buf, int send_buf_len)
  {
  int rslt;
  int err;

  if (inst_data->use_ssl == 0)
	{
	/* send clear text */
	if (inst_data->send_size == 0)
	  {
	  rslt = send(sock, send_buf, send_buf_len, 0);
	  show_send_data(inst_data, send_buf, send_buf_len);
	  if (rslt < 0)
		{
		sprintf(log_msg, "error sending data: %s", strerror(errno));
		tj_log(inst_data, log_msg, c_red);
		}
	  }
	else
	  {
	  /* buffer size set */
	  int loc, seg_len = inst_data->send_size;
	  for (loc = 0; loc < send_buf_len; loc += seg_len)
		{
		if ((send_buf_len - loc) < seg_len)
		  seg_len = send_buf_len - loc;
		rslt = send(sock, &send_buf[loc], seg_len, 0);
		show_send_data(inst_data, &send_buf[loc], seg_len);
		if (inst_data->send_delay > 0)
		  usleep(inst_data->send_delay);
		if (rslt < 0)
		  {
		  sprintf(log_msg, "error sending data: %s", strerror(errno));
		  tj_log(inst_data, log_msg, c_red);
		  break;
		  }
		}
	  }
	}
  else
	{
	/* send using SSL */
	if (inst_data->send_size == 0)
	  {
	  err = SSL_write(ssl, send_buf, send_buf_len);
	  show_send_data(inst_data, send_buf, send_buf_len);
	  if ((err) == -1)
		{
		sprintf(log_msg, "SSL write %s",
			ERR_error_string(ERR_get_error(), NULL));
		tj_log(inst_data, log_msg, c_red);
		}
	  }
	else
	  {
	  /* buffer size */
	  int loc, seg_len = inst_data->send_size;
	  for (loc = 0; loc < send_buf_len; loc += seg_len)
		{
		if ((send_buf_len - loc) < seg_len)
		  seg_len = send_buf_len - loc;
		err = SSL_write(ssl, &send_buf[loc], seg_len);
		show_send_data(inst_data, &send_buf[loc], seg_len);
		if (inst_data->send_delay > 0)
		  usleep(inst_data->send_delay);
		if ((err) == -1)
		  {
		  sprintf(log_msg, "SSL write %s", ERR_error_string(ERR_get_error(),
			  NULL));
		  tj_log(inst_data, log_msg, c_red);
		  }
		}
	  }
	}
  }

struct addrinfo *check_address(struct tj_data *inst_data, const char *addr,
	char *port, int passive)
  {

  /* check address */
  int n;
  struct addrinfo hints, *result;
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_flags = AI_CANONNAME;

  /* set socket type */
  hints.ai_socktype = SOCK_STREAM;

  /* set address family */
  if (inst_data->ipv6)
	hints.ai_family = AF_INET6;
  else
	hints.ai_family = AF_INET;

  if (passive)
	hints.ai_flags = AI_PASSIVE;

  if (inst_data->verbose && !passive)
	tj_log(inst_data, "resolving address...", c_black);

  if ((n = getaddrinfo(!addr[0] ? NULL : addr, port, &hints, &result)) != 0)
	{
	sprintf(log_msg, "invalid address: %s", gai_strerror(n));
	tj_log(inst_data, log_msg, c_red);
	if (inst_data->gui == 0)
	  exit(EXIT_FAILURE);
	else
	  return NULL;
	}
  else
	return result;

  }

int check_src_address(struct tj_data *inst_data, char *addr)
  {
  if (*addr == '\0')
	goto failure;

  /* check address */
  if (inst_data->verbose)
	tj_log(inst_data, "resolving source address...", c_black);
  struct hostent *hnames;
  hnames = gethostbyname(addr);
  if (!hnames)
	goto failure;
  else
	{
	inst_data->src_addr = hnames->h_addr_list[0];
	inst_data->src_addr_len = hnames->h_length;
	if (inst_data->verbose)
	  {
	  sprintf(log_msg, "setting source address to %s", inet_ntoa(
		  *(struct in_addr *) hnames->h_addr_list[0]));
	  tj_log(inst_data, log_msg, c_black);
	  }
	return 1;
	}

  failure:

  sprintf(log_msg, "invalid source address: %s", addr);
  tj_log(inst_data, log_msg, c_red);
  if (inst_data->gui == 0)
	exit(EXIT_FAILURE);
  else
	return 0;
  }

void rem_ipv6(struct tj_data *inst_data, char *bind_if_name, char *ip_addr)
  {
  int sock;
  struct in6_addr inaddr6;
  struct in6_ifreq ifr6;
  struct ifreq ifr;

  /* create socket */
  sock = socket(AF_INET6, SOCK_DGRAM, 0);
  if (sock < 0)
	{
	sprintf(log_msg, "error creating socket: %s", strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	goto done;
	}

  /* create address */
  if (inet_pton(AF_INET6, ip_addr, &inaddr6) <= 0)
	{
	sprintf(log_msg, "IP address error: %s", strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	goto done;
	}

  /* get interface */
  strcpy(ifr.ifr_name, bind_if_name);
  if (ioctl(sock, SIOGIFINDEX, &ifr) < 0)
	{
	sprintf(log_msg, "error removing IP address: %s", strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	goto done;
	}

  /* remove ip */
  ifr6.ifr6_addr = inaddr6;
  ifr6.ifr6_ifindex = ifr.ifr_ifindex;
  ifr6.ifr6_prefixlen = 64;
  if (ioctl(sock, SIOCDIFADDR, &ifr6) < 0)
	{
	sprintf(log_msg, "error removing IP address: %s", strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	}

  done:

  close(sock);
  return;
  }

int set_src_ip(struct tj_data *inst_data, char *bind_if_name, char *new_src_ip)
  {
  int sock;
  int error = 0;

  if (inst_data->ipv6)
	{
	/* IPv6 */
	static char last_src_ip[128] = "\0";
	if (last_src_ip[0])
	  {
	  if (strcmp(last_src_ip, new_src_ip))
		rem_ipv6(inst_data, bind_if_name, last_src_ip);
	  else
		return 0;
	  }

	struct in6_addr inaddr6;
	struct in6_ifreq ifr6;
	struct ifreq ifr;

	/* create socket */
	sock = socket(AF_INET6, SOCK_DGRAM, 0);
	if (sock < 0)
	  {
	  sprintf(log_msg, "error creating socket: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  error = -1;
	  goto done;
	  }

	/* create address */
	if (inet_pton(AF_INET6, new_src_ip, &inaddr6) < 1)
	  {
	  tj_log(inst_data, "IP address error", c_red);
	  error = -1;
	  goto done;
	  }

	/* get interface */
	strcpy(ifr.ifr_name, bind_if_name);
	if (ioctl(sock, SIOGIFINDEX, &ifr) < 0)
	  {
	  sprintf(log_msg, "error setting up interface: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  error = -1;
	  goto done;
	  }

	/* add ip */
	ifr6.ifr6_addr = inaddr6;
	ifr6.ifr6_ifindex = ifr.ifr_ifindex;
	ifr6.ifr6_prefixlen = 64;
	if (ioctl(sock, SIOCSIFADDR, &ifr6) < 0)
	  {
	  sprintf(log_msg, "error setting source IP address: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  error = -1;
	  goto done;
	  }
	strcpy(last_src_ip, new_src_ip);
	}
  else
	{
	/* IPv4 */
	struct in_addr inaddr;
	struct ifreq ifr;
	union
	  {
		struct sockaddr sa;
		struct sockaddr_in sa_in;
	  } s;

	/* create socket */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
	  {
	  sprintf(log_msg, "error creating socket: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  error = -1;
	  goto done;
	  }

	/* create address */
	if (inet_pton(AF_INET, new_src_ip, &inaddr) < 1)
	  {
	  tj_log(inst_data, "IP address error", c_red);
	  error = -1;
	  goto done;
	  }

	/* set interface */
	memset(&ifr, 0, sizeof(struct ifreq));
	strcpy(ifr.ifr_name, bind_if_name);

	/* add ip */
	s.sa_in.sin_family = AF_INET;
	s.sa_in.sin_addr = inaddr;
	ifr.ifr_addr = s.sa;
	if (ioctl(sock, SIOCSIFADDR, &ifr) < 0)
	  {
	  sprintf(log_msg, "error setting source IP address: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  error = -1;
	  }
	}

  done:

  close(sock);
  return error;

  }

